//
// numerical solution of the algebraic equation 
// p(z) = z(z-1)(z-2)(z-3)(z-4) = 0
// by the 2nd order Durand-Kerner method
//
// by Hidenori Ogata (27 October 2015)
//
#include <iostream>
#include <fstream>
#include <iomanip>
#include <math.h>
#include <complex> 
#define NSOL 10
#define NSTEP 100

using namespace std;

//
// polynomial p(z)
//
// n_sol : the degree of q(z) = the number of the roots
// a_coef : the coefficients of q(z)
//
complex<double> pol(int n_sol, double a_coef[NSOL+1], complex<double> z)
{
  complex<double> p;
  p = a_coef[0];
  for (int n=1; n<=n_sol; ++n) p = p * z + a_coef[n];
  //
  return p;
}
//
// each step of the Durand-Kerner method
//
// n_sol : the degree of q(z) = the number of the roots
// a_coef : the coefficients of q(z)
// z_sol[nu][j] : the j-th approximate root at the nu-th step
//
void dk_step(int n_sol, double a_coef[NSOL+1], 
	     complex<double> z_sol[NSTEP+1][NSOL+1], int nu)
{
  complex<double> z_old[NSOL+1], z_new[NSOL+1];
  for (int j=1; j<=n_sol; ++j) z_old[j] = z_sol[nu][j];
  //
  for (int j=1; j<=n_sol; ++j)
    {
      complex<double> denom;
      denom = 1.0;
      for (int k=1; k<=j-1; ++k) denom *= z_old[j] - z_old[k];
      for (int k=j+1; k<=n_sol; ++k) denom *= z_old[j] - z_old[k];
      denom *= a_coef[0];
      //
      z_new[j] = z_old[j] - pol(n_sol, a_coef, z_old[j]) / denom;
    }
  //
  for (int j=1; j<=n_sol; ++j) z_sol[nu+1][j] = z_new[j];
}
//
// convergence test
//
// n_sol : the degree of q(z) = the number of the roots
// a_coef : the coefficients of q(z)
// z_sol[nu][j] : the j-th approximate root at the nu-th step
// err_tol : error tolerance
// res_tol : residual tolerance
// l_conv : if convergent, return l_conv = 1
//          otherwise, return l_conv = 0
//
void conv_test(int n_sol, double a_coef[NSOL+1], 
	       complex<double> z_sol[NSTEP+1][NSOL+1], int nu, 
	       double err_tol, double res_tol, int & l_conv)
{
  l_conv = 0;
  //
  complex<double> z_old[NSOL+1], z_new[NSOL+1];
  //
  for (int j=1; j<=n_sol; ++j) {
    z_old[j] = z_sol[nu  ][j];
    z_new[j] = z_sol[nu+1][j];
  }
  //
  double dz_max, p_max;
  dz_max = 0.0;
  for (int j=1; j<=n_sol; ++j)
    {
      double dz_abs = abs(z_new[j] - z_old[j]);
      if (dz_abs > dz_max) dz_max = dz_abs;
    }
  p_max = 0.0;
  for (int j=1; j<=n_sol; ++j)
    {
      double p_abs = abs(pol(n_sol, a_coef, z_new[j]));
      if (p_abs > p_max) p_max = p_abs;
    }
  if (dz_max < err_tol || p_max < res_tol) l_conv = 1;
}
//
// iterations of the Durand-Kerner method
//
// n_sol : the degree of q(z) = the number of the roots
// a_coef : the coefficients of q(z)
// z_sol[nu][j] : the j-th approximate root at the nu-th step
// z_solution[j] : the j-th approximate root 
// l_conv : if convergent, return l_conv = 1
//          otherwise, return l_conv = 0
// nu_end : the number of iterations
//
void dk_method(int n_sol, double a_coef[NSOL+1], 
	       complex<double> z_sol[NSTEP+1][NSOL+1], 
	       complex<double> z_solution[NSOL+1], int & l_conv, int & nu_end)
{
  double eps = 1.0e-15;
  double err_tol, res_tol;
  err_tol = res_tol = eps;
  double z_max = 0.0;
  for (int j=1; j<=n_sol; ++j)
    {
      double z_abs = abs(z_sol[0][j]);
      if (z_abs > z_max) z_max = z_abs;
    }
  if (z_max!=0.0) err_tol *= z_max;
  double p_max = 0.0;
  for (int j=1; j<=n_sol; ++j)
    {
      double p_abs = abs(pol(n_sol, a_coef, z_sol[0][j]));
      if (p_abs > p_max) p_max = p_abs;
    }
  if (p_max!=0.0) res_tol *= p_max;
  //
  std::cout << std::setiosflags(std::ios::scientific);
  int nu_max = 20;
  //
  l_conv = 0;
  for (int nu=0; nu<=nu_max; ++nu)
    {
      dk_step(n_sol, a_coef, z_sol, nu);
      for (int j=1; j<=n_sol; ++j)
	cout << setw(2) << "z(" << nu+1 << ", " 
	     << setw(2) << j << " ) = " 
	     << setprecision(15) << z_sol[nu+1][j] << endl;
      cout << endl;
      conv_test(n_sol, a_coef, z_sol, nu, err_tol, res_tol, l_conv);
      if (l_conv==1) {
	for (int j=1; j<=n_sol; ++j) z_solution[j] = z_sol[nu+1][j];
	nu_end = nu+1;
	break;
      }
    }
  //
  if (l_conv==0) {
    nu_end = nu_max+1;
    for (int j=1; j<=n_sol; ++j) z_solution[j] = z_sol[nu_max+1][j];
  }
}
//
// generate the data for drawing the paths of the approximate solutions
// 
// n_sol : the number of the roots = the degree of p(z)
// z_sol[nu][j] : the j-th approximate solution of the nu-th step
// nu_end : the number of iterations 
//
void draw_graph(int n_sol, complex<double> z_sol[NSTEP+1][NSOL+1], int nu_end)
{
  ofstream ofs;
  ofs.open("graph.dat");
  for (int j=1; j<=n_sol; ++j)
    {
      for (int nu=0; nu<=nu_end; ++nu)
	{
	  complex<double> zj = z_sol[nu][j];
	  ofs << real(zj) << " " << imag(zj) << endl;
	}
      ofs << endl;
    }
  ofs.close();
}
//
// draw the circle on which the initial guesses are distributed
//
// n_sol : the nubmer of the roots = the degree of p(z)
// a_coef : the coefficients of p(z)
// radius : the radius of the circle
//
void draw_circle(int n_sol, double a_coef[NSOL+1], double radius)
{
  complex<double> I(0, 1);
  //
  ofstream ofs;
  ofs.open("circle.dat");
  //
  double beta = - a_coef[1] / (n_sol * a_coef[0]);
  double dt;
  int nt;
  nt = 256;
  dt = 2.0 * M_PI / nt;
  ofs << setprecision(3);
  for (int it=0; it<=nt; ++it)
    {
      double t = it * dt;
      complex<double> z = beta + radius * exp(I*t);
      ofs << real(z) << " " << imag(z) << endl;
    }
  ofs.close();
}
//
// the initial guess by Aberth
//
// n_sol : the number of the roots = the degree of p(z)
// a_coef : the coefficients of p(z)
// radius : the radius of the circle 
//          on which the initial solutions are distributed
// z0 : the initial guess
//
void initial_guess(int n_sol, double a_coef[NSOL+1], double radius, 
		   complex<double> z0[NSOL+1])
{
  complex<double> I(0, 1);
  //
  double beta;
  beta = - a_coef[1] / (n_sol * a_coef[0]);
  //
  for (int j=1; j<=n_sol; ++j)
    z0[j] = beta + radius * exp(I * ((2.0*M_PI*(j-1) + 1.5) / n_sol));
}
//
// the main program
//
int main (void)
{
  double a_coef[NSOL+1] = {1.0, -10.0, 35.0, -50.0, 24.0, 0.0};
  double radius;
  int n_sol;
  complex<double> z_sol[NSTEP+1][NSOL+1], z0[NSOL+1], z_solution[NSOL+1];
  //
  n_sol = 5;
  radius = 7.0;
  initial_guess(n_sol, a_coef, radius, z0);
  for (int j=1; j<=n_sol; ++j) z_sol[0][j] = z0[j];
  std::cout << std::setiosflags(std::ios::scientific);
  for (int j=1; j<=n_sol; ++j)
    cout << "z(" << setw(2) << 0 << ", " 
	 << setw(2) << j << ") = " 
	 << setprecision(15) << z_sol[0][j] << endl;
  cout << endl;
  //
  int l_conv, nu_end;
  dk_method(n_sol, a_coef, z_sol, z_solution, l_conv, nu_end);
  cout << endl;
  //
  if (l_conv==0) cout << "DK method not convergent." << endl;
  cout << "solution" << endl;
  for (int j=1; j<=n_sol; ++j)
    cout << setw(2) << "z(" << j << ") = " 
	 << setprecision(15) << z_solution[j] << endl;
  draw_graph(n_sol, z_sol, nu_end);
  draw_circle(n_sol, a_coef, radius);
  //
  return 0;
}
